#define NEW gsd_dummy_manager_new
#define START gsd_dummy_manager_start
#define STOP gsd_dummy_manager_stop
#define MANAGER GsdDummyManager
#include "gsd-dummy-manager.h"

#include "daemon-skeleton.h"
